/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

#include "eclib.h"

/* routines for ordered link list */

#define ZC_NEXTLINK(ptr, offset) *((char **)ptr+2*offset)
#define ZC_PREVLINK(ptr, offset) *((char **)ptr+2*offset + 1)

char *ZC_prevlink;
static char *ZC_nextlink;

ZC_o_listlink(list, t, offset)
char **list, *t;
{
if(!ZC_prevlink)
ZC_nextlink = ZC_NEXTLINK(list, offset);
else ZC_nextlink = ZC_NEXTLINK(ZC_prevlink, offset);
ZC_NEXTLINK(t, offset) = ZC_nextlink;
ZC_PREVLINK(t, offset) = ZC_prevlink;
if(ZC_nextlink)
ZC_PREVLINK(ZC_nextlink, offset) = t;
else ZC_PREVLINK(list, offset) = t;
if(ZC_prevlink)
ZC_NEXTLINK(ZC_prevlink, offset) = t;
else ZC_NEXTLINK(list, offset) = t;
} /* ZC_o_listlink */

ZC_o_listunlink(list, t, offset)
char **list, *t;
{
ZC_nextlink = ZC_NEXTLINK(t, offset);
ZC_prevlink = ZC_PREVLINK(t, offset);
ZC_NEXTLINK(t, offset) = 0;
ZC_PREVLINK(t, offset) = 0;
if(ZC_nextlink)
ZC_PREVLINK(ZC_nextlink, offset) = ZC_prevlink;
else ZC_PREVLINK(list, offset) = ZC_prevlink;
if(ZC_prevlink)
ZC_NEXTLINK(ZC_prevlink, offset) = ZC_nextlink;
else ZC_NEXTLINK(list, offset) = ZC_nextlink;
} /* ZC_o_listunlink */

/* insert the node at the head or tail of the list */
char *ZC_ht_insert(list, t, copy, struct_type, at_head)
char **list, *t;
{
if(!t) eclib_error(NODE_NULL_PTR, 1);
if(!list) eclib_error(NODE_NULL_PTR, 2);
if(copy) t = ZC_copy_struct(t, struct_type);

if(at_head)
ZC_prevlink = 0;
else ZC_prevlink = ZC_PREVLINK(list, 0);
ZC_o_listlink(list, t, 0);
/* increment number of nodes in this list */
list += 2;
++*((short *)list);
return t;
} /* ZC_ht_insert */

/* remove the node from the head or tail of the list */
char *ZC_ht_rm(list, at_head)
char **list;
{
char *t;
if(!list) eclib_error(NODE_NULL_PTR, 3);
if(at_head)
t = ZC_NEXTLINK(list, 0);
else t = ZC_PREVLINK(list, 0);
if(t) {
ZC_o_listunlink(list, t, 0);
/* decrement number of nodes in this list */
list += 2;
--*((short *)list);
}
return t;
} /* ZC_ht_rm */

ZC_extract(list, t, struct_type, drop)
char *t, **list;
{
int i;
if(!t) eclib_error(NODE_NULL_PTR, 4);
if(!list) eclib_error(NODE_NULL_PTR, 5);
for(i=0; i<struct_desc[struct_type].norders; ++i)
ZC_o_listunlink(list, t, i);
/* decrement number of nodes in this list */
list += 2*i;
--*((short *)list);
if(drop) ZC_free_struct(t, struct_type);
} /* ZC_extract */

/* assumes struct->next is the first element in a stack/chain/fifo/list node */
ZC_zap_list_gen(list, struct_type)
char **list;
{
char *t, *drop;
int i;

if(!list) eclib_error(NODE_NULL_PTR, 6);
if(struct_desc[struct_type].nodetype < 2)
eclib_error(NODE_STRUCT_TYPE, struct_type);

for(t = *list; t; ) {
drop = t;
t = * (char **)t;
ZC_free_struct(drop, struct_type);
} /* end freeing nodes in the chain */

/* finally, zero the list structure
 * don't run memset(list, 0, struct_size[struct_type+1]);
 * because the structure may not be allocated; it may just be a static
 * structure in memory with other variables immediately following it,
 * and the real structure size may be a couple bytes shorter than struct_size[]. */
for(i=0; i<struct_desc[struct_type].norders; ++i) {
*list++ = 0; /* head */
*list++ = 0; /* tail */
}
* (short *)list = 0; /* kill list->n_nodes */
} /* ZC_zap_list_gen */

/* copy a list, horribly inefficient */
ZC_copy_list_gen(dest, src, struct_type)
char **src, **dest;
{
register i;
int n, norders, index;
char **xref;
char **s1, **t1, *s2;

if(struct_desc[struct_type].nodetype == 0)
eclib_error(NODE_STRUCT_TYPE, struct_type);
if(struct_desc[struct_type].nodetype == 1) {
struct serial_inmem *x = (struct serial_inmem *)dest;
extern char **ZC_serial_grabblock();

memcpy(dest, src, sizeof(struct serial_inmem));
x->inmem_ptr = ZC_serial_grabblock(struct_desc[struct_type].size * x->n_room);
x->inmem = *x->inmem_ptr;
return;
}

norders = struct_desc[struct_type].norders * 2;
n = * ((short *)(src + norders));
xref = (char **) emalloc(n * 2 * sizeof(char *));
* ((short *) (dest + norders)) = n;

t1 = dest;
s1 = src;
i = 0;
while(*s1) {
*t1 = ZC_copy_struct(*s1, struct_type);
xref[i++] = *s1;
xref[i++] = *t1;
s1 = (char **) *s1;
t1 = (char **) *t1;
}
*t1 = 0;

/* now adjust the remaining links */
s1 = src;
t1 = dest;
while(s1) {
for(index=1; index<norders; ++index) {
s2 = s1[index];
if(!s2) t1[index] = 0;
else {
for(i=0; i<2*n; i+=2)
if(s2 == xref[i]) break;
if(i == 2*n) eclib_error(NODE_LIST_PTR, struct_type);
t1[index] = xref[i+1];
}
} /* next index */
s1 = (char **) *s1;
t1 = (char **) *t1;
} /* next node */

free(xref);
} /* ZC_copy_list_gen */

