/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* admin.c: run administrative commands */

#include <errno.h>

#include "structs.h"

run_admin_command(cmd)
char *cmd;
{
if(!strcmp(cmd, "stubgo")) { stubgo(); exit(0); }
if(!strcmp(cmd, "newid")) {
printf("If you have already transferred any of your data into another database\n\
for integration, changing your database ID can cause confusion.\n\
Consult the gng.doc file for more information on this `newid' command.\n");
if(get_yn("Shall I generate a new ID number?"))
dbidset();
exit(0);
}
if(!strcmp(cmd, "reindex")) { eval_index(); exit(0); }
if(!strcmp(cmd, "feedback")) { send_feedback(); exit(0); }
if(!strcmp(cmd, "printout")) { printgo(0); exit(0); }
if(!strcmp(cmd, "transmit")) { unload_db(1); exit(0); }
if(!strcmp(cmd, "unload")) { unload_db(0); exit(0); }
if(!strcmp(cmd, "load")) { load_db(0); exit(0); }
if(!strcmp(cmd, "foldin")) { load_db(1); exit(0); }
if(!strcmp(cmd, "textin")) { textin(); exit(0); }
if(!strcmp(cmd, "ver")) { printf("Version 2.2\n"); exit(0); }
eprint("usage:  gng [feedback|transmit|textin|unload|load|newid|ver|printout|foldin|reindex]");
} /* run_admin_command */

textin()
{
date textdate;
char filename[60];
char *archname;
int year, m, d;
FILE *fpin, *fpout;
char *s, etype;
char line[200];
long lineno = 1L, nentries = 0L;
static char badform[] = "line %ld of %s is not properly formatted.\n\
Text entries should be delimited by lines of the form  d>04/23,\n\
for a diary entry on April 23.";
extern long strtol();

set_curperson(searchfor(NULL));
printf("Enter file containing the ASCII text entries: ");
gets(filename);
if(!strcmp(filename, "q")) exit(0);
lifetime_dates(select_person(cur_prs_rowid), cur_fname, 0, 120);
do {
textdate = getdate("year of text entries", datemsg, 0);
} while(!textdate);
year = textdate / YEAR_DAY_RATIO;
archname = textfilename(cur_prs_rowid, textdate);
nl();

fpin = efopen(filename, "r");
fpout = NULL;
while(fgets(line, 200, fpin)) {
if(line[1] != '>') { /* just another line of text */
if(!fpout)
eprint(badform, lineno, filename);
fputs(line, fpout);
} else {

if(fpout) {
fclose(fpout);
textin_finish(archname, textdate, etype);
fpout = NULL;
}

textdate = 0;
etype = tolower(line[0]);
if(strchr("dcpi", etype)) {
m = strtol(line+2, &s, 10);
if(m && *s++ == '/') {
d = strtol(s, NULL, 10);
textdate = datecv(m, d, year);
}
}
if(!textdate)
eprint(badform, lineno, filename);
fputs(line, stdout);
fpout = efopen(temp_textfile, "w");
++nentries;
}

if(line[strlen(line)-1] == '\n')
++lineno;
} /* loop over the archive file */

if(fpout) {
fclose(fpout);
textin_finish(archname, textdate, etype);
fpout = NULL;
}

fclose(fpin);
bump_te_count((int)nentries);
unlink(temp_textfile);

printf("\n%ld text entries folded in\n", nentries);
} /* textin */

textin_finish(archname, textdate, etype)
char *archname;
long textdate;
char etype;
{
long l, textin_text();

l = textin_text(archname, textdate, etype);
if(l >= 0)
bloblink(cur_prs_rowid, textdate, l, this_dbid, etype);
} /* textin_finish */

send_feedback()
{
FILE *f;

f = efopen(temp_textfile, "w");
fprintf(f, "Subject: gng_msg %ld feedback\n\n", this_dbid);
fclose(f);

user_edit();
send_email();
} /* send_feedback */

mailstring(msgtype, s)
char *msgtype, *s;
{
FILE *f;
f = efopen(temp_textfile, "w");
fprintf(f, "Subject: gng_msg %ld %s\n%s\n", this_dbid, msgtype, s);
fclose(f);
send_email();
} /* mailstring */

send_email()
{
#ifdef MSDOS
eprint("Sorry, you can only send email from a UNIX machine with internet access");
#else
FILE *f;
int len;
char email_addr[100];
char mailcmd[100];

email_addr[0] = 0;
f = fopen("data/email", "r");
if(!f && errno == EMFILE) {
uncache_fd();
f = fopen("data/email", "r");
}
if(f) {
if(fgets(email_addr, 80, f)) {
len = strlen(email_addr);
if(email_addr[len-1] == '\n')
email_addr[len-1] = 0;
}
fclose(f);
}

if(strcmp(email_addr, "no-mail")) {
if(!email_addr[0])
eprint("Canot find internet address, missing or unreadable file gng/data/email.\n\
Please create this file and insert the line %s",
CURRENT_EMAIL);

sprintf(mailcmd, "mail %s <%s", email_addr, temp_textfile);
if(system(mailcmd))
eprint("could not execute the command:\n%s", mailcmd);
}

unlink(temp_textfile);
#endif
} /* send_email */

setup_email()
{
#ifndef MSDOS
FILE *f;

f = efopen("data/email", "w");
if(get_yn("Does your machine have access to internet email?"))
fprintf(f, "%s\n", CURRENT_EMAIL);
else
fprintf(f, "no-mail\n");
fclose(f);

mailstring("init", "Database initialized");
#endif
} /* setup_email */

unload_db(outgoing)
{
FILE *f1, *f2;
struct IDX_DESC *ib;
char withtext;
int i, fd1;

if(db_is_empty()) {
printf("The database is empty, canot unload\n");
return;
}

withtext = get_yn("Shall I unload text entries as well?");
if(withtext)
withtext += get_yn("Does this include personal text entries (i.e. diary entries and letters)?");

fd1 = ecreat("unld");
for(i=0; i<ZC_max_struct; ++i) {
if(i == STRUCT_textent) continue;
if(i == STRUCT_address) continue;
rel_unload(fd1, i);
}
close(fd1);

f2 = efopen("unld2", "w");
fd1 = ebinopen("unld");
f1 = fdopen(fd1, "r");
if(outgoing) {
fprintf(f2, "Subject: gng_msg %ld transmit\n", this_dbid);
}
fprintf(f2, "dbid%ld\n", this_dbid);
fprintf(f2, "ratios:");
for(ib=idx_desc; ib->cls; ++ib)
fprintf(f2, " %d/%d", ib->numer, ib->denom);
fprintf(f2, "\nver%d\n", this_version);
uuenc(f1, f2);
fprintf(f2, "end data\n");
fclose(f1);

if(!withtext) {
fprintf(f2, "notext\n");
} else {
long prsid, strtol();
char *fn;

fprintf(f2, "%s\n", withtext == 2 ? "textall" : "text");

printf("Unloading text entries.\n");
dirstart = 1;
while(fn = gng_nextfile()) {
/* deduce person from filename */
prsid = strtol(fn+5, NULL, 10);

/* read through file and extract text entries */
fd1 = ebinopen(fn);
unld_text(fd1, f2, prsid, withtext-1);
close(fd1);
} /* loop over files */
} /* doing text entries */
fclose(f2);
erename("unld2", "unld");

if(outgoing) { /* transmit data */
erename("unld", temp_textfile);
send_email();
unlink(temp_textfile);
}
} /* unload_db */

/* ENC is the basic 1 character encoding function to make a char printing */
#define ENC(c) (((c) & 077) + ' ')
/* single character decode */
#define DEC(c)	(((c) - ' ') & 077)

uuenc(in, out)
FILE *in, *out;
{
	char buf[80];
	int i, n;

	for (;;) {
		/* 1 (up to) 45 character line */
		n = fr(in, buf, 45);
		eputc(ENC(n), out);

		for (i=0; i<n; i += 3)
			outdec1(&buf[i], out);

		eputc('\n', out);
		if (n <= 0)
			break;
	}
} /* uuenc */

uudec(in, out)
FILE *in;
FILE *out;
{
	char buf[80];
	char *bp;
	int n;

	for (;;) {
		/* for each input line */
		if (fgets(buf, sizeof buf, in) == NULL)
eprint("unexpected EOF in uudecode()");
		n = DEC(buf[0]);
		if (n <= 0)
			break;

		bp = &buf[1];
		while (n > 0) {
			outdec2(bp, out, n);
			bp += 4;
			n -= 3;
		}
	}
} /* uudec */

outdec1(p, f)
char *p;
FILE *f;
{
	int c1, c2, c3, c4;

	c1 = *p >> 2;
	c2 = (*p << 4) & 060 | (p[1] >> 4) & 017;
	c3 = (p[1] << 2) & 074 | (p[2] >> 6) & 03;
	c4 = p[2] & 077;
	eputc(ENC(c1), f);
	eputc(ENC(c2), f);
	eputc(ENC(c3), f);
	eputc(ENC(c4), f);
} /* outdec1 */

outdec2(p, f, n)
char *p;
FILE *f;
{
	int c1, c2, c3;

	c1 = DEC(*p) << 2 | DEC(p[1]) >> 4;
	c2 = DEC(p[1]) << 4 | DEC(p[2]) >> 2;
	c3 = DEC(p[2]) << 6 | DEC(p[3]);
	if (n >= 1)
		eputc(c1, f);
	if (n >= 2)
		eputc(c2, f);
	if (n >= 3)
		eputc(c3, f);
} /* outdec2 */

fr(fd, buf, cnt)
FILE *fd;
char *buf;
int cnt;
{
	int c, i;

	for (i=0; i<cnt; i++) {
		c = getc(fd);
		if (c == EOF)
			return(i);
		buf[i] = c;
	}
	return  cnt;
} /* fr */

grabline(f)
FILE *f;
{
int l;
if(!fgets(line, 256, f))
eprint("unexpected EOF reading load file");
/* in case the Unix file was not undos-ed */
l = strlen(line);
if(l >= 2 && line[l-2] == '\r' && line[l-1] == '\n') {
line[l-2] = '\n';
line[l-1] = 0;
}
} /* grabline */

static char mapnum_file[] = "mapnum";
int mfd;
long foldincount;

mapserial(lptr)
long *lptr;
{
elseek(mfd, 4*(*lptr), 0);
eread(mfd, lptr, 4);
} /* mapserial */

static prsid_shift(withtext)
{
int ifd, ofd;
int len, relnum;
long prsid;
struct person *p;
char t[100];

printf("Building map between source serial numbers and internal serial numbers.\n");

ifd = ebinopen("data/logfile");
ofd = ecreat("data/logfile1");
mfd = eopen(mapnum_file, O_BINARY|O_RDWR|O_TRUNC|O_CREAT, 0666);
foldincount = 0;

while(eread(ifd, t, 8)) {
relnum = ((short *)t)[0];
len = ((short *)t)[1];
if(relnum >= ZC_max_struct || struct_desc[relnum].size != len)
eprint("corrupt logfile, relation %d length %d", relnum, len);
prsid = ((long*)t)[1];
((long*)t)[1] = 0;

eread(ifd, t+8, len);
if(relnum != STRUCT_person) {
mapserial((long*)(t+8));
if(relnum == STRUCT_prslink)
mapserial((long*)(t+12));
ewrite(ofd, t, len+8);
continue;
}

file_extend(mfd, prsid*4, 0L);
elseek(mfd, prsid*4, 0);
if(withtext < 2) {
p = (struct person *) (t+8);
if(withtext == 1) /* cannot trust the count */
p->txtc1 = 0;
else { /* no text entries, count is 0 */
p->txtc1 = 1;
p->txtcnt = 0;
}
}
prsid = insupd_person(NULLKEY, t+8);
ewrite(mfd, &prsid, 4);
++foldincount;
} /* reading log tupples */

close(ofd);
close(ifd);
erename("data/logfile1", "data/logfile");
} /* prsid_shift */

load_db(foldin)
{
FILE *f, *fl;
struct IDX_DESC *ib;
char ratset = 0, *s;
int fd;
long vnum, dbid, strtol();
char withtext;
short ver;
char *fn;

fd = ebinopen("ld");
f = fdopen(fd, "r");
grabline(f);
if(strncmp(line, "dbid", 4) ||
!isdigit(line[4]))
eprint("load file does not begin with DB identification number");
dbid = strtol(line+4, NULL, 10);

/* if index line is present and we're doing a load, rebuild ratios */
grabline(f);
if(!strncmp(line, "ratios", 6)) {
ratset = 1;
ib = idx_desc;
s = line;
while(s=strchr(s, ' ')) {
++s;
if(ib->cls) {
if(!foldin)
sscanf(s, "%d/%d", &ib->numer, &ib->denom);
++ib;
}
}
grabline(f);
}

if(strncmp(line, "ver", 3) ||
!isdigit(line[3]))
eprint("load file does not contain a version number");
vnum = strtol(line+3, NULL, 10);
ver = vnum;
if(ver != this_version)
eprint("Data contained in your loadfile is version %d, while your program is\n\
version %d.  Cannot load incompatible data",
ver, this_version);

if(!foldin) {
if(!db_is_empty()) {
if(dbid == this_dbid)
printf("This command overwrites the current instance of your database\n\
with the instance encoded in the load file `ld'.\n\
Any data not contained in this load file will be lost.\n");
else
printf("This command replaces your database with another database, containing a\n\
different internal identification number.  All your data will be overwritten.\n");
if(!get_yn("Do you wish to procede?")) return;
}
} else {
if(dbid == this_dbid)
eprint("The data you are attempting to incorporate comes from a database\n\
having the same identification number as your own.  In short, it looks like\n\
you are trying to fold in extra copies of your own data.\n\
Please make sure your unique database identification number is different from\n\
any other database that you might be sharing data with.");
}

if(!foldin) {
serial_db_drop();
serial_db_create();
if(ratset) idxval_write();

dirstart = 1;
while(fn = gng_nextfile())
unlink(fn);
}

/* uudecode the data to build the logfile */
fd = ecreat("data/logfile");
fl = fdopen(fd, "w");
uudec(f, fl);
fclose(fl);
grabline(f);
if(strncmp(line, "end data", 8))
eprint("no end line after uudecode()");
printf("Restore file constructed.\n");

grabline(f);
withtext = 1;
if(!strncmp(line, "notext", 6)) withtext = 0;
else {
if(strncmp(line, "text", 4))
eprint("line after `end data' should read `text' or `notext'");
if(!strncmp(line, "textall", 7)) withtext = 2;
}

if(foldin)
prsid_shift(withtext);

printf("Beginning load procedure, please wait.\n");
printf("Load procedure may be slow; only a couple hundred people per minute.\n");
serial_restore();
unlink("data/logfile");

if(withtext) {
/* now for the text entries */
printf("Loading text entries.\n");
ld_text(f, foldin);
}
fclose(f);
nl();

if(!foldin) prs_txtcounts(withtext);


if(!foldin && dbid != this_dbid) {
fd = ecreat("data/dbid");
ewrite(fd, &dbid, 4);
close(fd);
printf("\nIf this database is a copy of the original (and both will continue to expand),\n\
rather than a simple move to another computer, you may need to run `gng newid'.\n\
Consult the gng.doc file for more information.\n");
}

if(foldin) {
printf("%ld people folded into your database.\n", foldincount);
if(get_yn("Would you like me to try to merge some of these new people with the\n\
individuals that are already present in your database?\n\
i.e.  Have you seen some of these people before?"))
merge_search();
close(mfd);
unlink(mapnum_file);
}
} /* load_db */

prs_txtcounts(withtext)
{
struct rel_cursor c;
struct person p, *p1;

if(withtext == 2) return; /* counts are ok */

LOOP_person(&c, p1, &p)
if(withtext == 1) /* cannot trust the count */
p.txtc1 = 0;
else {
p.txtc1 = 1;
p.txtcnt = 0;
}
insupd_person(c.rowid, &p);
END_LOOP_person(&c)
} /* prs_txtcounts */

