/***********************************************************************
|  Genealogy management system
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* text_d.c: connect text files to the database */

#include "structs.h"

char *textfilename(prsid, textdate)
long prsid;
date textdate;
{
static char filename[30];
int year;

if(textdate) {
year = textdate / YEAR_DAY_RATIO;
year %= 100;
sprintf(filename, "text/%ld.Y%d", prsid, year);
} else sprintf(filename, "text/%ld.Y__", prsid);
return filename;
} /* textfilename */

/* how many text entries, and what was the first year */
textstats(prsid, ecount, firstyear)
long prsid;
int *ecount, *firstyear;
{
struct rel_cursor c;
struct textent *te;
long l;
int cnt;

/* find the first entry for this person */
l = 9999*YEAR_DAY_RATIO;
cnt = 0;
LOOP_textent(&c, te, NULL)
if(te->prsid != prsid) continue;
++cnt;
if(te->when && te->when < l) l = te->when;
END_LOOP_textent(&c)

*firstyear = l/YEAR_DAY_RATIO;
*ecount = cnt;
} /* textstats */

bloblink(prsid, textdate, offset, source, etype)
long prsid, offset, source;
date textdate;
char etype;
{
struct textent te;

init_0(&te);
te.source = source;
te.offset = offset;
te.prsid = prsid;
te.when = textdate;
te.etype = etype;
insupd_textent(0L, &te);
} /* bloblink */

bump_te_count(diff)
unsigned diff;
{
cur_prs.txtcnt += diff;
insupd_person(cur_prs_rowid, &cur_prs);
} /* bump_te_count */

text_interactive()
{
struct rel_cursor cursor;
date textdate, newdate;
char c, cmd;
long l;
char lrs[12], *filename;
struct textent *te, new_te;
int i, year;
char *s;

textstats(cur_prs_rowid, &i, &year);
printf("\n%d text entries for %s", i, cur_fname);
if(year < 9999)
printf(", dated entries begin in the year %d\n", year);
nl();

if(cur_prs.txtc1 == 0 || cur_prs.txtcnt != i) {
cur_prs.txtc1 = 1;
cur_prs.txtcnt = i;
insupd_person(cur_prs_rowid, &cur_prs);
}

top1:
nl();
lifetime_dates(&cur_prs, cur_fname, 0, 120);
textdate = getdate("date of text entries", datemsg, 0);
year = textdate/YEAR_DAY_RATIO;
filename = textfilename(cur_prs_rowid, textdate);

top2:
spreadnum = 0;
LOOP_textent(&cursor, te, NULL)
if(te->prsid != cur_prs_rowid) continue;
if(te->when < textdate) continue;
if(te->when / YEAR_DAY_RATIO != year) continue;
spread_insert(cursor.rowid, cur_prs_rowid, te->when, 0);
END_LOOP_textent(&cursor)

printf("\n%d text entries gathered for %s\n", spreadnum, cur_fname);

top3:
if(spreadnum) {
printf("\n    1     2     3     4     5     6     7     8     9     0  ");
for(i=0; i<spreadnum; ++i) {
if(!(i%10))
printf("\n%c", (spreadnum <= 10) ? ' ' : i/10 + 'a');
te = select_textent(spreaddata[i].rownum);
if(spreaddata[i].order) {
s = datestring(spreaddata[i].order);
strcpy(s+2, s+3);
s[4] = 0;
} else s = "----";
printf(" %c%s", toupper(te->etype), s);
} /* end loop displaying entries */
nl();
} /* entries present */

printf("\nHit s to select a new date, a to add a new entry");
if(spreadnum) {
if(spreadnum == 50)
printf(", n for the next page");
printf(",\n    r to review an entry, or u to update an entry");
}
cmd = promptfor(":", spreadnum ? "sanru" : "sa");

if(cmd == 's') { nl(); goto top1; }

if(cmd == 'n') {
nl();
textdate = spreaddata[spreadnum-1].order;
if(textdate && spreadnum < SPREADMAX) ++textdate;
goto top2;
}

if(cmd == 'a') {
nl();
if(year)
newdate = getdate("date of text entry", datemsg, year);
else {
newdate = 0;
if(spreadnum == SPREADMAX) {
printf("Cannot have more than %d undated text entries\n", SPREADMAX);
goto top3;
}
}
c = promptfor("type of entry [diary,correspondence,information,publication]", "cdip");
nl();

init_0(&new_te);
new_te.etype = c;
new_te.prsid = cur_prs_rowid;
new_te.when = newdate;

l = edfile(filename, -1L, 'u', newdate, c);
if(l >= 0) {
new_te.offset = l;
new_te.source = this_dbid;
insupd_textent(NULLKEY, &new_te);
bump_te_count(1);
}
unlink(temp_textfile);
goto top2;
} /* end adding entry */

/* review or update an entry */
i = 0;
if(spreadnum > 10) {
strcpy(lrs, "abcde");
lrs[(spreadnum-1)/10+1] = 0;
c = promptfor("   row", lrs);
i = 10*(c-'a');
}

strcpy(lrs, "1234567890");
if(spreadnum-i < 10) lrs[spreadnum-i] = 0;
if(cmd == 'r') strcat(lrs, "*");
c = promptfor("   column", lrs);
nl();
if(c == '*') {
char ehead[30];
int fdin, fdout;
long last_update;

fdout = ecreat(temp_10file);

for(c=0; c<10; ++c,++i) {
if(i == spreadnum) break;
te = select_textent(spreaddata[i].rownum);
#ifdef MSDOS
sprintf(ehead, "\r\n%c>%s:\r\n\r\n",
toupper(te->etype), te->when ?  datestring(te->when) : "--/--");
strcpy(ehead+9, ":\r\n\r\n");
ewrite(fdout, ehead, 14);
#else
sprintf(ehead, "\n%c>%s:\n\n",
toupper(te->etype), te->when ? datestring(te->when) : "--/--");
strcpy(ehead+8, ":\n\n");
ewrite(fdout, ehead, 11);
#endif

extract_file(filename, te->offset);
fdin = ebinopen(temp_textfile);
l = filesize(fdin);
elseek(fdin, 0L, 0);
if(copyfile(fdin, fdout, l))
eprint("error copying %s to %s", temp_textfile, temp_10file);
close(fdin);
} /* end loop over files in the row */
close(fdout);

erename(temp_10file, temp_textfile);

fdin = ebinopen(temp_textfile);
last_update = moddate(fdin);
close(fdin);

user_edit();

fdin = ebinopen(temp_textfile);
if(last_update != moddate(fdin))
printf("\nCannot update multiple text entries simultaneously.\n\
Any changes you made to the text will not be recorded.\n\n");
close(fdin);
unlink(temp_textfile);
goto top3;
} /* end '*' processing */

c -= '0';
if(c == 0) c = 10;
i += c-1;

select_copy_textent(spreaddata[i].rownum, &new_te);
l = edfile(filename, new_te.offset, cmd, new_te.when, new_te.etype);
if(l < 0) {
delete_textent(spreaddata[i].rownum);
bump_te_count(-1);
unlink(temp_textfile);
goto top2;
}

if(l != new_te.offset) {
new_te.offset = l;
insupd_textent(spreaddata[i].rownum, &new_te);
}
unlink(temp_textfile);
goto top3;
} /* text_interactive */

up_shortnote()
{
int i;
long rownums[7];
struct qknote qn;
char ntype;

printf("Short note categories are birth, death, medical, education, occupation,\n\
religion, and ancestry.  Please use the initial letter.\n\n");

if(sn_display(stdout, rownums)) nl();

while(1) {
ntype = promptfor("Enter note type, or `.' when you are finished", qkcat);
nl();
i = strchr(qkcat, ntype) - qkcat;

getqknote(ntype);
if(line[0]) {
init_0(&qn);
qn.ntype = ntype;
qn.prsid = cur_prs_rowid;
strcpy(qn.note, line);
rownums[i] = insupd_qknote(rownums[i], &qn);
} else {
if(rownums[i]) delete_qknote(rownums[i]);
rownums[i] = 0;
}

if(ntype == 'b')
set_curperson(cur_prs_rowid);
}
} /* up_shortnote */

up_name()
{
struct rel_cursor c;
struct name *n, upname;
int i;
char cmd, lrs[12];

/* gather and display names for update */
spreadnum = 0;
IDX_LOOP_name(&c, n, NULL, cur_prs_rowid, "prsid")
/* birth name comes first, give it a low (though nonzero) year */
if(n->birth) n->start = EARLY_DATE;
spread_insert(c.rowid, cur_prs_rowid, n->start, 0);
END_LOOP_name(&c)

for(i=0; i<spreadnum; ++i) {
if(i == 9) {
printf("  Only the first 9 names are accessible\n");
break;
}

select_copy_name(spreaddata[i].rownum, &upname);
splitstring(upname.fullname);
printf("%d %s %s %s\n", i+1, fld1, fld2, fld3);
}
nl();

strcpy(lrs, "a123456789");
lrs[i+1] = 0;
cmd = promptfor("Enter a digit to update an existing name or a to add a new name:", lrs);
nl();

lifetime_dates(&cur_prs, cur_fname, 0, 120);
n = getname("full name", 0, 0);

if(cmd == 'a') {
n->start = getdate("date this name became valid", datemsg, 0);
n->prsid = cur_prs_rowid;
insupd_name(NULLKEY, n);
return;
} /* add name */

i = cmd - '0' - 1;
select_copy_name(spreaddata[i].rownum, &upname);
strcpy(upname.fullname, n->fullname);
upname.hashval = n->hashval;
if(!upname.birth) {
if(upname.start) {
printf("Name has a start date of %s\n", datestring(upname.start));
if(get_yn("Do you wish to change this start date?"))
upname.start = getdate("new start date", datemsg, 0);
} else {
upname.start = getdate("date this name became valid", datemsg, 0);
}
}
insupd_name(spreaddata[i].rownum, &upname);
set_curperson(cur_prs_rowid);
} /* up_name */

up_date()
{
date d;
char change;

printf("Simply hit return for the dates you do not wish to change\n");
change = 0;
if(cur_prs.birth)
printf("birthdate is %s.  ", datestring(cur_prs.birth));
lowdate = highdate = 0;
d = getdate("birthdate", datemsg, 0);
if(d) { cur_prs.birth = d; change = 1; }
if(cur_prs.birth) {
if(cur_prs.death)
printf("deathdate is %s.  ", datestring(cur_prs.death));
lifetime_dates(&cur_prs, cur_fname, 0, 120);
/* so we can make the death date later */
highdate = lowdate + 120 * YEAR_DAY_RATIO;
d = getdate("deathdate", datemsg, 0);
if(d) { cur_prs.death = d; change = 1; }
}

if(change)
insupd_person(cur_prs_rowid, &cur_prs);

/* now for the marriage dates */
linkscan_date((cur_prs.sex == 'm') ? 'w' : 'h');
} /* up_date */

