/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* str_help.c: help with dates and names and strings */

#include <errno.h>

#include "eclib.h"

static short ndays[] = { 0,31,29,31,30,31,30,31,31,30,31,30,31};

char *datestring(packed)
date packed;
{
static char s[11];
int i;
short y = packed/YEAR_DAY_RATIO;
short d = packed%YEAR_DAY_RATIO;

if(!y) return "???";

if(d) {
ndays[2] = 29;
for(i=1; i<=12; ++i) {
if(d <= ndays[i]) break;
d -= ndays[i];
}
sprintf(s, "%02d/%02d/%04d", i, d, y);
} else {
sprintf(s, "%04d", y);
}
return s;
} /* datestring */

date datecv(m, d, y)
{
int i;
if(m < 1 || m > 12) { printf("month %d out of range\n", m); return 0L; }
ndays[2] = 29;
if(!isleapyear(y))
ndays[2] = 28;
if(d < 1 || d > ndays[m]) { printf("day %d out of range\n", d); return 0L; }
ndays[2] = 29;
for(i=1; i<m; ++i)
d += ndays[i];
return y*YEAR_DAY_RATIO + d;
} /* datecv */

/* convert string into date.
 * Format is 08/31/1963, or  4 17 1875. */
date string_to_date(s, lowyear, highyear)
char *s;
{
int y, m, d;
int ok = 1;
date retdate;

m = strtol(s, &s, 10);
if(*s == '/' || *s == ' ')
d = strtol(++s, &s, 10);
else ok = 0;
if(*s == '/' || *s == ' ')
y = strtol(++s, &s, 10);
else ok = 0;
if(*s) ok = 0; /* extra stuff */
if(m < 1 || m > 12 ||
d < 1 || d > ndays[m] ||
d == 29 && m == 2 && !isleapyear(y) ||
y < lowyear || y > highyear)
ok = 0;

retdate = 0;
if(ok) retdate = datecv(m,d,y);
return retdate;
} /* string_to_date */

isleapyear(y)
int y;
{
if(y%4 == 0) { /* ok so far */
if(y%100 || y%400 == 0)
return 1;
}
return 0;
} /* isleapyear */

/* add n days to a date.  Not very efficient. */
date add_days(d, n)
date d;
int n;
{
int y, leftover;
int check_y;
int sign;

if(!n) return d;

y = d/YEAR_DAY_RATIO;
leftover = d%YEAR_DAY_RATIO;
if(!leftover) eprint("cannot add days to a year-only date");

if(leftover == 31+29) {
++leftover;
++d;
--n;
if(!n) return d;
} /* don't start on leapday */

sign = 1;
if(n < 0) n = -n, sign = -1;

while(n >= 183) {
n-=365;
check_y = y;
y += sign;
/* did this jump in year cross a leapday? */
if(sign > 0 && leftover > 31+29) check_y = y;
if(sign < 0 && leftover < 31+29) check_y = y;
if(isleapyear(check_y)) --n; /* extra day lost */
} /* end while jumping by years */

if(n < 0) n = -n, sign = -sign;

while(n > 0) {
leftover += sign;
--n;
if(leftover == 31+29 && !isleapyear(y))
leftover += sign;
if(leftover == 0)
leftover = 365, --y;
if(leftover == 367)
leftover = 1, ++y;
} /* end adding days */

d = (date)y*YEAR_DAY_RATIO + leftover;
return d;
} /* add_days */

/* subtract two dates.  Not very efficient */
diff_dates(d1, d2)
date d1, d2;
{
int n = 0, sign, sign1;
int check_y;
int y1, y2;
int l1, l2;
date d3;
int l3;

sign = 1;
if(d2 > d1) sign = -1, d3 = d1, d1 = d2, d2 = d3;

y1 = d1/YEAR_DAY_RATIO;
l1 = d1%YEAR_DAY_RATIO;
y2 = d2/YEAR_DAY_RATIO;
l2 = d2%YEAR_DAY_RATIO;
if(!l1 || !l2)
eprint("cannot subtract year-only dates");
if(l1 == 31+29) {
if(l2 > 1 && l2 != 31+29+1) --l1, --l2;
else
++l1, ++l2;
}

while(y1 > y2) {
check_y = y1;
--y1;
n += 365;
if(l1 < 31+29) check_y = y1;
if(isleapyear(check_y)) ++n;
} /* years are now equal */

sign1 = 1;
if(l2 > l1) sign1 = -1, l3 = l1, l1 = l2, l2 = l3;
l3 = l1 - l2;
if(l1 > 31+29 && l2 < 31+29 && !isleapyear(y1)) --l3;
n += sign1*l3;
return n*sign;
} /* diff_dates */

/* difference in days does not easily translate into years,
 * especially if you need to know for certain whether somebody
 * is or is not 65, or 18.
 * Call diff_years(today, birthdate); */
diff_years(d1, d2)
date d1, d2;
{
int y1 = d1/YEAR_DAY_RATIO;
int l1 = d1%YEAR_DAY_RATIO;
int y2 = d2/YEAR_DAY_RATIO;
int l2 = d2%YEAR_DAY_RATIO;
if(!l1 || !l2)
eprint("cannot subtract year-only dates");
y1 -= y2;
if(l1 < l2) --y1;
return y1;
} /* diff_years */

/* convert date to weekday.
 * Simplest possible implementation given the above routines */
date_wd(d)
date d;
{
int ndays;
if(!(d%YEAR_DAY_RATIO))
eprint("cannot find weekday for year-only date");
ndays = diff_dates(d, YEAR_DAY_RATIO*1970L+4);
ndays %= 7;
if(ndays < 0) ndays += 7;
return ndays;
} /* date_wd */

/* shift string to upper, lower, or mixed case */
smartmix(s, action)
char *s;
{
char c, *t;
int mix, last_lr;

mix = 0;
for(t=s; c = *t; ++t) {
if(islower(c)) {
mix |= 1;
if(action == 1) *t = toupper(c);
}
if(isupper(c)) {
mix |= 2;
if(action == -1) *t = tolower(c);
}
} /* end chec for both upper and lower */

if(action || mix == 3) return;

/* no case specified, make our best guess */
last_lr = 0;
for(t=s; c = *t; ++t) {
if(isalpha(c)) {
if(last_lr) c = tolower(c);
else c = toupper(c);
*t = c;
last_lr = 1;

/* check for McDonld */
if(t-s >= 1 && c == 'c' && t[-1] == 'M')
last_lr = 0;
} else last_lr = 0;
} /* end setting case */
} /* smartmix */

/* case independent string compare.
 * n = 0 means compare the entire strings. */
ci_strncmp(s, t, n)
char *s, *t;
int n;
{
int i = 0;
char c1, c2;

for(; *s | *t; ++s, ++t) {
if(n) { /* maybe we are done */
if(i++ == n) return 0;
}
if(!*s) return -1;
if(!*t) return 1;
c1 = tolower(*s);
c2 = tolower(*t);
c1 -= c2;
if(c1) return c1;
}
return 0;
} /* ci_strncmp */

/* compress white space */
spacecrunch(s)
char *s;
{
int i, j;
char c, space = 1;

for(i=j=0; c = s[i]; ++i) {
if(isspace(c)) {
if(!space) s[j++] = ' ', space = 1;
continue;
}
s[j++] = c, space = 0;
}
if(space && j) --j; /* drop trailing space */
s[j] = 0;
} /* spacecrunch */

/* shift string one to the right */
shiftright(s, first)
char first, *s;
{
int l = strlen(s);
for(; l>=0; --l)
s[l+1] = s[l];
s[0] = first;
} /* shiftright */

/* shift string one to the left */
shiftleft(s)
char *s;
{
strcpy(s, s+1);
} /* shiftleft */

bell() { printf("\07\b"); }
nl() { printf("\n"); }
dot() { printf("."); fflush(stdout); }

#ifndef MSDOS
/* simulate MSDOS getche() system call */
#ifdef compile_trouble
getche()
{
char buf[20];
gets(buf);
return buf[0];
} /* getche */
#else
#include <termio.h>

getche()
{
int c;
char lastmin, lasteol;
struct termio buf;

c = ioctl(0, TCGETA, &buf);
if(c < 0) eprint("cannot use ioctl to run termio in raw mode");
/* leave echo on */
buf.c_lflag &= ~ICANON;
lastmin = buf.c_cc[VMIN];
lasteol = buf.c_cc[VEOL];
buf.c_cc[VMIN] = 1;
buf.c_cc[VEOL] = 0;
ioctl(0, TCSETAW, &buf);
c = getchar();
buf.c_lflag |= ICANON;
buf.c_cc[VMIN] = lastmin;
buf.c_cc[VEOL] = lasteol;
ioctl(0, TCSETAW, &buf);
return c;
} /* getche */
#endif
#endif

get_no_echo()
{
int c;

top:
c = getche();
if(c== EOF) exit(1);
/* control-C = break */
if(c== 3) exit(1);
#ifdef MSDOS
if(!c) { /* function or alt key */
c = getche();
bell();
goto top;
}
#endif
return c;
} /* get_no_echo */

/* change to a base directory and stay their throughout the running program */
cd_basedir(name)
char *name;
{
char evname[10]; /* environment variable */
char *s, *t, *getenv();

strcpy(evname, name);
strcat(evname, "dir");
smartmix(evname, 1); /* upshift */

s = t = getenv(evname);
if(!s || !*s)
s = getenv("HOME");
if(!s || !*s)
s = "/";
if(chdir(s))
eprint("cannot cd to base directory %s", s);

if(!t || !*t) {
if(chdir(name)) {
/* let's try making the directory */
mkdir(name, 0777);
if(chdir(name))
eprint("1cannot access or create directory %s%s%s",
s, strcmp(s,"/") ? "/" : "", name);
}
}
} /* cd_basedir */

int scanner_base, err_base;

/* a better version of strtol() */
zc_strtol(s, cptr, lptr)
char *s, **cptr;
long *lptr;
{
int base = scanner_base;
char badnum = 0;
char c = *s;
long l = 0;

if(c == '0' && !base) {
base = 8;
c = *++s;
if(toupper(c) == 'X') { base = 16; ++s; }
}
if(!base) base = 10;
for(; isalnum(c = *s); ++s) {
if(badnum) continue;
if(isalpha(c)) c = toupper(c) - ('A' - '9' - 1);
c -= '0';
if(c >= base) {
badnum = -1;
*cptr = s;
err_base = base;
continue;
}
if(l > (unsigned long)0xffffffff/base ||
(l *= base) > (unsigned long)0xffffffff - c) {
badnum = -2;
*cptr = s;
continue;
}
l += c;
}
*lptr = l;
if(!badnum) *cptr = s;
return badnum;
} /* zc_strtol */

#define SS_CACHE 50

char *stringspace = 0;
unsigned short ss_room = 400;
short ss_length;

char *code_to_string(n)
unsigned short n;
{
char *s;
if(!stringspace || n >= ss_length || (s = stringspace+n)[-1])
eprint("@cannot convert %u into a string, length %u", n, ss_length);
return s;
} /* code_to_string */

str_code
string_to_code(s)
char *s;
{
static char ss_head, ss_tail, ss_next[SS_CACHE], ss_last[SS_CACHE], ss_used;
static unsigned short ss_word[SS_CACHE];
static char k64[] = "@space taken up by strings consumes more than 64K";
unsigned short len, n;
char c, *t;
extern char *malloc(), *realloc();

len = strlen(s) + 1;

/* first call sets things up */
if(!stringspace) {
if(len+2 > ss_room) ss_room = len+2;
stringspace = malloc(ss_room);
if(!stringspace) eprint("@malloc failure initializing stringspace[%u]", ss_room);
stringspace[0] = stringspace[1] = 0;
ss_length = 2;
/* only the empty string in cache */
ss_word[0] = 1;
ss_head = ss_tail = 0;
ss_next[0] = ss_last[0] = -1;
ss_used = 1;
} /* end setting things up */

/* search through cache */
for(c=ss_head; c>=0; c=ss_next[c]) {
n = ss_word[c];
t = stringspace + n;
if(*t != *s) continue;
if(!*t) goto match_c;
if(*++t != s[1]) continue;
if(!*t) goto match_c;
if(strcmp(++t, s+2)) continue;
match_c:
/* move this word to the head */
if(c != ss_head) { /* we have work to do */
char cl = ss_last[c];
char cn = ss_next[c];
ss_next[cl] = cn;
if(cn >= 0) ss_last[cn] = cl;
else ss_tail = cl;
ss_last[ss_head] = c;
ss_next[c] = ss_head;
ss_head = c;
} /* moving this word to the head of the list */
return n;
} /* end search through cache */

/* now look through all names */
for(t=stringspace+1; t<stringspace+ss_length; t += strlen(t)+1)
if(!strcmp(t, s)) break;
if(t > stringspace + ss_length)
eprint("@string space misalignment %u/%u", t-s, ss_length);

if(t == stringspace + ss_length) { /* name not found */
if(len > (unsigned short)0xffff - ss_length)
eprint(k64);
if(ss_length + len > ss_room) {
if(ss_room < 40000)
ss_room = ss_room / 2 * 3;
else ss_room = (unsigned short)0xffff;
if(ss_length + len > ss_room) ss_room = ss_length + len;
stringspace = realloc(stringspace, ss_room);
if(!stringspace)
eprint("@malloc failure expanding stringspace[%u]", ss_room);
} /* end reallocating stringspace */

t = stringspace + ss_length;
strcpy(t, s);
ss_length += len;
} /* new name added */

n = t - stringspace; /* get offset */
/* put this at the head of the cache */
if(ss_used < SS_CACHE) {
ss_last[ss_head] = ss_used;
ss_next[ss_used] = ss_head;
ss_last[ss_used] = -1;
ss_head = ss_used;
ss_word[ss_used++] = n;
} else { /* free up tail */
c = ss_tail;
ss_tail = ss_last[c];
ss_next[ss_tail] = -1;
ss_last[c] = -1;
ss_next[c] = ss_head;
ss_last[ss_head] = c;
ss_head = c;
}
return n;
} /* string_to_code */

#ifdef MSDOS
#include <dos.h>
#else
/* unfortunately reading a directory is highly and unavoidably unportable.
 * If all else fails, replace the block of code with
 * the less efficient   popen("ls text/*.*", "r"). */
#include <dirent.h>
#endif

int dirstart;

/* loop through the files in a directory */
char *nextfile(base)
char *base;
{
char *s, *t;
static char nextfilename[50];
#ifdef MSDOS
static struct WORDREGS regs;
static char dta[43]; /* disk transfer address for directory search */

/* set up disk transfer address */
regs.ax = 0x1a00;
regs.dx = (unsigned) dta;
intdos(&regs, &regs);

top:
if(dirstart) {
regs.ax = 0x4e00;
sprintf(nextfilename, "%s/*.*", base);
regs.dx = (unsigned) nextfilename;
regs.cx = 0x11; /* ignore hidden files */
dirstart = 0;
} else {
regs.ax = 0x4f00;
regs.dx = (unsigned) dta;
}
intdos(&regs, &regs);

 if(regs.cflag & 1) return 0;
s = dta+30;
#else
static DIR *df;
static struct dirent *de;

if(dirstart) {
df = opendir(base);
if(!df && errno == EMFILE && ZC_fn_uncache) {
(*ZC_fn_uncache)();
df = opendir(base);
}
if(!df) eprint("cannot open %s directory", base);
dirstart = 0;
}

top:
de = readdir(df);
if(!de) { closedir(df); return 0; }
/* if your dirent.h hasn't got d_ino defined, try d_fileno */
if(de->d_ino == 0) goto top;
s = de->d_name;
#endif
if(!isdigit(s[0])) goto top;
t = strchr(s, '.');
if(!t || t[1] != 'Y') goto top;

sprintf(nextfilename, "%s/%s", base, s);
return nextfilename;
} /* nextfile */

